{*************************************************************************}
{*   TarFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\ArchiveFormats\Tar\TarFormat.pas                       *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 5th November 2003  (05.11.2003)                            }
{*************************************************************************}

{This unit contains class for accessing TAR file archives.}

unit TarFormat;

{$I 'Compiler.inc'}

interface

uses
  Kernel, Objects, Platforms, Files, TarArchive;

type

  {This archive format supports extracting files from Tar archives.}
  TNTarFormat = class(TNArchiveFormat)
  protected
    procedure DoReadFile(Archive: TNArchive; const FileName: string; var Data:
      Pointer; var Size: Long); override;
    procedure DoWriteFile(Archive: TNArchive; const FileName: string; Data:
      Pointer; Size: Long); override;
    function DoOpen(const FileName: string): TNArchive; override;
    procedure DoClose(Archive: TNArchive); override;
    procedure DoFillFileList(Archive: TNArchive); override;
  public
    constructor Create; override;

    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sTarExtensions = 'tar';
  sTarFormatName = 'Tar Archive';

var

  TarInstance: TNTarFormat = nil;


{*************************************************************************}
{                         IMPLEMEMTATION                                  }
{*************************************************************************}


implementation

uses
  SysUtils;


{*************************************************************************}
{                        TNTarFormat methods                              }
{*************************************************************************}


constructor TNTarFormat.Create;
begin
  inherited Create;
  FExtensions := sTarExtensions;
  FFormatName := sTarFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNTarFormat.DoReadFile(Archive: TNArchive; const FileName: string; var
  Data: Pointer; var Size: LongInt);
begin
  TNTarArchive(Archive.Handle).ReadFile(SwapPathDelims(FileName, '/'), Data,
    Size);
end;

procedure TNTarFormat.DoWriteFile(Archive: TNArchive; const FileName: string;
  Data: Pointer; Size: Long);
var
  S: string;
begin
  // if file is in subdirectory, add subdirectories too. It works
  // without it, but may be somewhere somehow ...
  S := SwapPathDelims(FileName, PathDelim);
  S := ExtractFileDir(S);
  while S <> '' do
  begin
    TNTarArchive(Archive.Handle).WriteDir(SwapPathDelims(S, '/'));
    S := ExtractFileDir(S);
  end;
  TNTarArchive(Archive.Handle).WriteFile(SwapPathDelims(FileName,'/'), Data, Size);
end;

function TNTarFormat.DoOpen(const FileName: string): TNArchive;
var
  Arc: TNTarArchive;
begin
  Result := nil;
  Arc := TNTarArchive.Create(FileName);
  if Arc = nil then
    Exit;
  Result := TNArchive.Create(Self, Arc);
  Result.Path := FileName;
  Result.Name := ExtractFileName(FileName);
end;

procedure TNTarFormat.DoClose(Archive: TNArchive);
begin
  TNTarArchive(Archive.Handle).Free;
end;

procedure TNTarFormat.DoFillFileList(Archive: TNArchive);
begin
  inherited DoFillFileList(Archive);
  TNTarArchive(Archive.Handle).GetFileList(Archive.Files);
end;

function TNTarFormat.TestFormat(Stream: TNStream): Boolean;
var
  OldPos: Int64;
  Magic: Char;
begin
  Result := False;
  if Stream = nil then Exit;
  OldPos := Stream.Position;
  if Stream.Size < 512 then Exit;
  Stream.Read(Magic, SizeOf(Magic));
  Result := Ord(Magic) > 31;
  Stream.Position := OldPos;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}

initialization

  TarInstance := TNTarFormat.Create;

finalization

  FreeAndNil(TarInstance);

end.

