{*************************************************************************}
{*   Files                                                               *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\Files.pas                                              *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 5th November 2003  (05.11.2003)                            }
{*************************************************************************}

{This unit contains classes and functions working with files,
 file archives, compression, ...}

unit Files;

{DONE -oGalfar : Add ZipArchive today!}
{DONE -oGalfar : Find out better way how to register things.
Make list with reg/unreg functions and override Find(Ext) for
TNModule descendant types. Make acess to register by function.}
{TODO -oGalfar: Look at CrystalSpace's VFS mounting - vfs.cfg}
{DONE -oGalfar: Add writing to archive Tar !! rewrite LibTar to suite my needs!
 overwriting,..}
{TODO -oGalfar: Find something between rle anz zlib - like Hafman, but working.}
{TODO -oGalfar: Add GZ Compress internal}
{DONE -oGalfar: Make cosmetic changes in ArcFormat and Compression like in
ImageFIleFormat. DoOpen,DoClose,...}
{TODO -oGalfar: Add some virtual file system - put to another file,
  delete TNFileManager}

{$I 'Compiler.inc'}

interface

uses
  Kernel, Objects, Platforms, SysUtils;


{*************************************************************************}
{                         Consts, Types                                   }
{*************************************************************************}


const

  {Some defalut directories.}
  {Directory for stroring various app settings.}
  SettingsDir = 'Settings';
  {Directory for stroring app data.}
  DataDir = 'Data';
  {Directory for stroring app graphic data.}
  GraphicDir = DataDir + PathDelim + 'Graphic';
  {Directory for stroring app music data.}
  MusicDir = DataDir + PathDelim + 'Music';
  {Directory for stroring app sound data.}
  SoundDir = DataDir + PathDelim + 'Sound';


  {Identifier used to differ Compression module library
   from another module libraries.}
  CmpLibraryID = 'cmp';
  {Identifier used to differ ArchiveFormat module library
   from another module libraries.}
  ArcLibraryID = 'arc';


{*************************************************************************}
{                         Class Types                                     }
{*************************************************************************}


type

  {This module compresses and decompresses input data.}
  TNCompression = class(TNFileModule)
  protected
    {These methods are called by Compress/Decompress methods to
     do real compression/decompression. Override these in descendants.}
    procedure DoCompress(InData: Pointer; InSize: Long; var OutData: Pointer; var
      OutSize: Long); virtual;
    procedure DoDecompress(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long); virtual;
  public
    {Compresses input data to output data.}
    function Compress(InData: Pointer; InSize: Long; var OutData: Pointer; var
      OutSize: Long): Boolean;
    {Decompresses input data to output data.}
    function Decompress(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long): Boolean;
  end;

  {Class refernce for TNCompression.}
  TNCompressionClass = class of TNCompression;

  TNArchive = class;

  {This module is abstract wrapper to APIs that control access to
   various archive file formats like zip, tar, ...}
  TNArchiveFormat = class(TNFileModule)
  protected
    {These methods are called by public methods to
     do real things. Override these in descendants.}
    procedure DoReadFile(Archive: TNArchive; const FileName: string; var Data:
      Pointer; var Size: Long); virtual;
    procedure DoWriteFile(Archive: TNArchive; const FileName: string; Data:
      Pointer; Size: Long); virtual;
    function DoOpen(const FileName: string): TNArchive; virtual;
    procedure DoClose(Archive: TNArchive); virtual;
    procedure DoFillFileList(Archive: TNArchive); virtual;
  public
    {Opens archive of given name and creates new instance
     of TNarchive for it. If opening fail, nil is returned.}
    function Open(const FileName: string): TNArchive;
    {Closes given archive. It is called in destructor of TNArchive.
     It does not free archive instance. Rather call Archive.Free.}
    procedure Close(Archive: TNArchive);
    {Fills archive's file list with actual contents of archive.}
    procedure FillFileList(Archive: TNArchive);
    {Returns True if given file is in Archive.}
    function IsFileInArchive(Archive: TNArchive; const FileName: string):
      Boolean; virtual;
    {Reads file (in Archive) of given name to Data buffer. Returns True if
     successful.}
    function ReadFile(Archive: TNArchive; const FileName: string; var Data:
      Pointer; var Size: Long): Boolean;
    {Writes file (to Archive) of given name contained in Data buffer. Returns
     True if successful. If file already exists in archive - it is overwritten.
     If archive format can not overwrite existing files nothing is done.}
    function WriteFile(Archive: TNArchive; const FileName: string; Data:
      Pointer; Size: Long): Boolean;
  end;

  {Class refernce for TNArchiveFormat.}
  TNArchiveFormatClass = class of TNArchiveFormat;

  {This class represents single file archive of any type. Internal format
   of archive (zip, tar, ...) is accessed trough Format property.}
  TNArchive = class(TNNamedPersistent)
  protected
    FFiles: TNSortedStrings;
    FFormat: TNArchiveFormat;
    FHandle: Pointer;
    FPath: string;
  public
    constructor Create(AFormat: TNArchiveFormat; AHandle: Pointer); reintroduce;
    destructor Destroy; override;

    {Returns True if given file is in archive.}
    function IsFileInArchive(const FileName: string): Boolean;
    {Reads file of given name to Data buffer. Returns True if successful.}
    function ReadFile(const FileName: string; var Data: Pointer; var Size:
      LongInt): Boolean; overload;
    {Writes file of given name contained in Data buffer. Returns True if successful.}
    function WriteFile(const FileName: string; Data: Pointer; Size: LongInt):
      Boolean;

    {Archive low-level handle used by archive formats.}
    property Handle: Pointer read FHandle;
    {File format of this archive (zip,tar,...).}
    property Format: TNArchiveFormat read FFormat;
  published
    {Contains list of all files in archive.}
    property Files: TNSortedStrings read FFiles write FFiles;
    {Full path to archive file.}
    property Path: string read FPath write FPath;
  end;

{$IFNDEF MODLIBRARY}
  TNFileManager = class(TNSingleton)
  private

  protected

  public
    constructor Create; override;
    destructor Destroy; override;
  published

  end;
{$ENDIF}


{*************************************************************************}
{                           Misc functions                                }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

{* TNCompression registration/other *}

{Registers new Compression instance so they can be used by system.}
procedure RegisterCompression(Format: TNCompression);
{Registers new Compression instances.}
procedure RegisterCompressions(Formats: array of TNCompression);
{UnRegisters Compression instance.}
procedure UnRegisterCompression(Format: TNCompression);
{UnRegisters new Compression instance.}
procedure UnRegisterCompressions(Formats: array of TNCompression);
{Returns Compression registered with given extension. If not found
 it returns nil.}
function FindCompression(const Ext: string): TNCompression;

{* TNArchiveFormat registration/other *}

{Registers new Archive File Format instance so they can be used by system.}
procedure RegisterArchiveFormat(Format: TNArchiveFormat);
{Registers new Archive File Format instances.}
procedure RegisterArchiveFormats(Formats: array of TNArchiveFormat);
{UnRegisters Archive File Format instance.}
procedure UnRegisterArchiveFormat(Format: TNArchiveFormat);
{UnRegisters new Archive File Format instance.}
procedure UnRegisterArchiveFormats(Formats: array of TNArchiveFormat);
{Returns Image file format registered with given extension. If not found
 it returns nil.}
function FindArchiveFormat(const Ext: string): TNArchiveFormat;

{$ENDIF}


{*************************************************************************}
{                         IMPLEMEMTATION                                  }
{*************************************************************************}


implementation

{$IFNDEF MODLIBRARY}
uses
  {Default archive file formats.}
  ZipFormat, TarFormat,
  {Default archive file formats.}
  ZLibFormat, RleFormat;
{$ENDIF}

var
  {List for holding reistered compressions formats. Accessed trough
   RegisteredCompressions function.}
  RegCompressions: TNFileModuleList = nil;

  {List for holding reistered archive file formats. Accessed trough
   RegisteredArchiveFormat function.}
  RegArchiveFormats: TNFileModuleList = nil;


{*************************************************************************}
{                          TNCompression methods                          }
{*************************************************************************}


procedure TNCompression.DoCompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long);
begin
end;

procedure TNCompression.DoDecompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long);
begin
end;

function TNCompression.Compress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long): Boolean;
begin
  Result := False;
  if FCanSave then
  try
    DoCompress(InData, InSize, OutData, OutSize);
    Result := True;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'Compress',
        'Error when compressing data! Data can be corrupted.'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
      Result := False;
    end;
  end;
end;

function TNCompression.Decompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long): Boolean;
begin
  Result := False;
  if FCanLoad then
  try
    DoDecompress(InData, InSize, OutData, OutSize);
    Result := True;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'Decompress',
        'Error when decompressing data! Data can be corrupted.'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
      Result := False;
    end;
  end;
end;


{*************************************************************************}
{                        TNArchiveFormat methods                          }
{*************************************************************************}


procedure TNArchiveFormat.DoReadFile(Archive: TNArchive;
  const FileName: string; var Data: Pointer; var Size: Integer);
begin
end;

procedure TNArchiveFormat.DoWriteFile(Archive: TNArchive;
  const FileName: string; Data: Pointer; Size: Integer);
begin
end;

procedure TNArchiveFormat.DoFillFileList(Archive: TNArchive);
begin
end;

function TNArchiveFormat.DoOpen(const FileName: string): TNArchive;
begin
  Result := nil;
end;

procedure TNArchiveFormat.DoClose(Archive: TNArchive);
begin
end;

function TNArchiveFormat.Open(const FileName: string): TNArchive;
begin
  try
    if not FileExists(FileName) then
      raise Exception.Create('File can not be found!');
    Result := DoOpen(FileName);
    FillFileList(Result);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'Open', 'Error when opening archive "'
        + FileName + '" !' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
      Result := nil;
    end;
  end;
end;

procedure TNArchiveFormat.Close(Archive: TNArchive);
begin
  if Archive <> nil then
  try
    DoClose(Archive);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Close', 'Error when closing archive "'
        + Archive.Name + '" !' + sLineBreak + sExceptMsg + E.Message),
        E.ClassName);
  end;
end;

procedure TNArchiveFormat.FillFileList(Archive: TNArchive);
var
  I: Long;
begin
  if Archive <> nil then
  try
    Archive.Files.Clear;
    DoFillFileList(Archive);
    Archive.Files.Sorted := False;
    for I := 0 to Archive.Files.Count - 1 do
      Archive.Files[I] := SwapPathDelims(Archive.Files[I], PathDelim);
    Archive.Files.Sorted := True;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'FillFileList', 'Error when filling file ' +
        'list for archive "' + Archive.Name + '" !' + sLineBreak + sExceptMsg +
        E.Message), E.ClassName);
  end;
end;

function TNArchiveFormat.IsFileInArchive(Archive: TNArchive; const FileName:
  string): Boolean;
var
  I: Long;
  OnlyName: string;
begin
  Result := False;
  // try search with Win and Linux path delimiters
  if Archive.Files.Find(SwapPathDelims(FileName, '\'), I) or
    Archive.Files.Find(SwapPathDelims(FileName, '/'), I) then
  begin
    Result := True;
    Exit;
  end;
  // try search in different directories
  OnlyName := ExtractFileName(FileName);
  for I := 0 to Archive.Files.Count - 1 do
    if SameStr(OnlyName, ExtractFileName(Archive.Files[I])) then
    begin
      Result := True;
      Exit;
    end;
end;

function TNArchiveFormat.ReadFile(Archive: TNArchive;
  const FileName: string; var Data: Pointer; var Size: Integer): Boolean;
begin
  Result := False;
  if FCanLoad and (Archive <> nil) then
  try
    DoReadFile(Archive, FileName, Data, Size);
    Result := True;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'ReadFile',
        'Error while reading file from archive!.' + sLineBreak +
        sExceptMsg + E.Message), E.ClassName);
      Result := False;
    end;
  end;
end;

function TNArchiveFormat.WriteFile(Archive: TNArchive;
  const FileName: string; Data: Pointer; Size: Integer): Boolean;
begin
  Result := False;
  if FCanSave and (Archive <> nil) then
  try
    DoWriteFile(Archive, FileName, Data, Size);
    FillFileList(Archive);
    Result := True;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'WriteFile',
        'Error while writing file to archive!.' + sLineBreak +
        sExceptMsg + E.Message), E.ClassName);
      Result := False;
    end;
  end;
end;


{*************************************************************************}
{                        TNArchive methods                                }
{*************************************************************************}


constructor TNArchive.Create(AFormat: TNArchiveFormat; AHandle: Pointer);
begin
  inherited Create;
  FHandle := AHandle;
  FFormat := AFormat;
  FFiles := TNSortedStrings.Create;
  if FFormat = nil then
    ShowError(ErrorStr(Self, 'Create', 'Format of this archive is nil!' +
      'If you do anything else with this archive, Access Violations will arise!'));
end;

destructor TNArchive.Destroy;
begin
  FFormat.Close(Self);
  FFormat := nil;
  FFiles.Free;
  inherited Destroy;
end;

function TNArchive.IsFileInArchive(const FileName: string): Boolean;
begin
  Result := FFormat.IsFileInArchive(Self, FileName);
end;

function TNArchive.ReadFile(const FileName: string; var Data: Pointer; var Size:
  LongInt): Boolean;
begin
  Result := FFormat.ReadFile(Self, FileName, Data, Size);
end;

function TNArchive.WriteFile(const FileName: string; Data: Pointer; Size:
  LongInt): Boolean;
begin
  Result := FFormat.WriteFile(Self, FileName, Data, Size);
end;


{*************************************************************************}
{                        TNFileManager methods                            }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

constructor TNFileManager.Create;
begin
  inherited Create;
end;

destructor TNFileManager.Destroy;
begin
  inherited Destroy;
end;

{$ENDIF}


{*************************************************************************}
{                           Misc functions                                }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

function RegisteredCompressions: TNFileModuleList;
begin
  if RegCompressions = nil then RegCompressions := TNFileModuleList.Create;
  Result := RegCompressions;
end;

procedure RegisterCompression(Format: TNCompression);
begin
  RegisteredCompressions.Add(Format);
end;

procedure RegisterCompressions(Formats: array of TNCompression);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    RegisterCompression(Formats[I]);
end;

procedure UnRegisterCompression(Format: TNCompression);
begin
  RegisteredCompressions.Remove(Format);
end;

procedure UnRegisterCompressions(Formats: array of TNCompression);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    UnRegisterCompression(Formats[I]);
end;

function FindCompression(const Ext: string): TNCompression;
begin
  Result := TNCompression(RegisteredCompressions.Find(Ext));
end;

function RegisteredArchiveFormats: TNFileModuleList;
begin
  if RegArchiveFormats = nil then RegArchiveFormats := TNFileModuleList.Create;
  Result := RegArchiveFormats;
end;

procedure RegisterArchiveFormat(Format: TNArchiveFormat);
begin
  RegisteredArchiveFormats.Add(Format);
end;

procedure RegisterArchiveFormats(Formats: array of TNArchiveFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    RegisterArchiveFormat(Formats[I]);
end;

procedure UnRegisterArchiveFormat(Format: TNArchiveFormat);
begin
  RegisteredArchiveFormats.Remove(Format);
end;

procedure UnRegisterArchiveFormats(Formats: array of TNArchiveFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    UnRegisterArchiveFormat(Formats[I]);
end;

function FindArchiveFormat(const Ext: string): TNArchiveFormat;
begin
  Result := TNArchiveFormat(RegisteredArchiveFormats.Find(Ext));
end;

{$ENDIF}


{*************************************************************************}
{                         Internall Stuff                                 }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

procedure LoadedCmp(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNCompressionClass;
  Inst: TNCompression;
  Check: Boolean;
begin
  if Entry^.ID = CmpLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      Check := True;
      for J := 0 to RegisteredCompressions.Modules.Count - 1 do
        if TNCompression(
          RegisteredCompressions.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Check := False;
          Break;
        end;
      if Check then
      begin
        Inst := Cls.Create;
        if Inst <> nil then
          RegisterCompressions(Inst);
      end;
    end;
  except
    raise Exception.Create('Error when creating Compression from ' +
      'module library!');
  end;
end;

procedure UnLoadedCmp(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNCompressionClass;
  Inst: TNCompression;
begin
  if Entry^.ID = CmpLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      for J := 0 to RegisteredCompressions.Modules.Count - 1 do
        if TNCompression(
          RegisteredCompressions.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Inst := TNCompression(RegisteredCompressions.Modules[J]);
          UnRegisterCompression(Inst);
          Inst.Free;
          Break;
        end;
    end;
  except
    raise Exception.Create('Error when unloading Compression from ' +
      'module library!');
  end;
end;

procedure LoadedArc(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNArchiveFormatClass;
  Inst: TNArchiveFormat;
  Check: Boolean;
begin
  if Entry^.ID = ArcLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      Check := True;
      for J := 0 to RegisteredArchiveFormats.Modules.Count - 1 do
        if TNArchiveFormat(
          RegisteredArchiveFormats.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Check := False;
          Break;
        end;
      if Check then
      begin
        Inst := Cls.Create;
        if Inst <> nil then
          RegisterArchiveFormat(Inst);
      end;
    end;
  except
    raise Exception.Create('Error when creating ArchiveFormat from ' +
      'module library!');
  end;
end;

procedure UnLoadedArc(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNArchiveFormatClass;
  Inst: TNArchiveFormat;
begin
  if Entry^.ID = ArcLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      for J := 0 to RegisteredArchiveFormats.Modules.Count - 1 do
        if TNArchiveFormat(
          RegisteredArchiveFormats.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Inst := TNArchiveFormat(RegisteredArchiveFormats.Modules[J]);
          UnRegisterArchiveFormat(Inst);
          Inst.Free;
          Break;
        end;
    end;
  except
    raise Exception.Create('Error when unloading ArchiveFormat from ' +
      'module library!');
  end;
end;

{$ENDIF}


initialization

{$IFNDEF MODLIBRARY}
  {Adds load/unload callbacks for ModuleRegister.}
  ModuleRegister.AddLoadCallBack(LoadedCmp);
  ModuleRegister.AddUnLoadCallBack(UnLoadedCmp);
  ModuleRegister.AddLoadCallBack(LoadedArc);
  ModuleRegister.AddUnLoadCallBack(UnLoadedArc);

  {Registration of default compression formats is put here instead of
   their original units to be sure that they will be registered in
   this sequence (first will be chosen as default format).}
  RegisterCompression(ZLibFormat.ZLibInstance);
  RegisterCompression(RleFormat.RleInstance);

  {Registration of default archive formats is put here instead of
   their original units to be sure that they will be registered in
   this sequence (first will be chosen as default format).}
  RegisterArchiveFormat(ZipFormat.ZIPInstance);
  RegisterArchiveFormat(TarFormat.TARInstance);
{$ENDIF}

finalization

  FreeAndNil(RegCompressions);
  FreeAndNil(RegArchiveFormats);

end.

