{*************************************************************************}
{*   CPUX86Info                                                          *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\CPUX86Info.pas                              *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 18th September 2003  (18.09.2003)                          }
{*************************************************************************}

{This unit contains functions to retrieve information about
installed CPU. It contains functions which can be on X86
based systems.}

unit CPUX86Info;

{$I 'Compiler.inc'}

interface

uses
  Kernel, Platforms;

{Returns information on installed CPU.  Remember that
it does NOT RETRIEVE ALL info - ProcessorCount and Frequency
are not filled, you must find then by platform specific things.}
procedure GetX86CPUInfo(var Info: TCPUInfo);


{*************************************************************************}
{                         IMPLEMENTATION                                  }
{*************************************************************************}


implementation

uses
  SysUtils;

const
  IDAMD = 'AuthenticAMD';
  IDIntel = 'GenuineIntel';
  IDCyrix = 'CyrixInstead';
  IDCentaur = 'CentaurHauls';

  sAMD = 'AMD';
  sIntel = 'Intel';
  sCyrix = 'Cyrix';
  sCentaur = 'Centaur/IDT';

  VAMD = 1;
  VIntel = 2;
  VCyrix = 3;
  VCentaur = 4;

procedure GetX86CPUInfo(var Info: TCPUInfo);
type
  TStrBuf = array[0..11] of Char;

  function HasCPUID: LongBool;
  asm
    PUSHFD
    POP  EAX
    MOV  ECX,EAX
    XOR  EAX,$00200000
    PUSH EAX
    POPFD
    PUSHFD
    POP  EAX
    XOR  EAX,ECX
  end;

  procedure CPUID(Flag: ULong; var Signature, Features: ULong);
  asm
    PUSH EBX
    PUSH ESI
    PUSH EDI
    MOV  ESI,EDX
    MOV  EDI,ECX
    DB   $0F,$A2
    MOV  [ESI],EAX
    MOV  [EDI],EDX
    POP  EDI
    POP  ESI
    POP  EBX
  end;

  function GetCPUVendor: string;

    function GetBuf: TStrBuf;
    asm
  	  PUSH    EBX
      PUSH    EDI
     	MOV     EDI,EAX
    	MOV     EAX,0
  	  DW      $A20F
    	MOV     EAX,EBX
    	XCHG	  EBX,ECX
    	MOV	    ECX,4
      @1:
    	STOSB
    	SHR     EAX,8
    	LOOP    @1
    	MOV     EAX,EDX
    	MOV	    ECX,4
      @2:
    	STOSB
    	SHR     EAX,8
     	LOOP    @2
    	MOV     EAX,EBX
    	MOV	    ECX,4
      @3:
    	STOSB
    	SHR     EAX,8
    	LOOP    @3
    	POP     EDI
	    POP     EBX
    end;

  begin
    Result := GetBuf;
  end;

  function MakeVendor(const ID: string): string;
  begin
    if ID = IDAMD then
      Result := sAMD
    else
      if ID = IDIntel then
        Result := sIntel
      else
        if ID = IDCyrix then
          Result := sCyrix
        else
          if ID = IDCentaur then
            Result := sCentaur
          else
            Result := ID;
  end;

  procedure GetStringAMD(Flag: ULong; Str: PChar);
  var
    RetVal, PS: ULong;
  asm
    PUSH  EBX
    MOV   PS,EDX
  	CPUID
    MOV	  RETVAL,EAX
   	MOV	  EAX, PS
  	MOV	  DWORD PTR[EAX+4],EBX
  	MOV	  DWORD PTR[EAX+8],ECX
  	MOV	  EBX,RETVAL
  	MOV	  DWORD PTR[EAX+12],EDX
  	MOV	  DWORD PTR[EAX],EBX
    POP   EBX
  end;

  function GetVendorNum(const Vendor: string): Long;
  begin
    if Vendor = sAMD then
      Result := VAMD
    else
      if Vendor = sIntel then
        Result := VIntel
      else
        if Vendor = sCyrix then
          Result := VCyrix
        else
          if Vendor = sCentaur then
            Result := VCentaur
          else
            Result := 0;
  end;

  function GetCPUName: string;
  begin
    case GetVendorNum(Info.Vendor) of
      VAMD:
        case Info.Family of
          5:
            case Info.Model of
              0: Result := 'K5 PR-75/90/100)';
              1: Result := 'K5 PR-120/133';
              2: Result := 'K5 PR-166';
              3: Result := 'K5 PR-200';
              6: Result := 'K6 0.3 micron';
              7: Result := 'K6 0.25 micron';
              8: Result := 'K6-2';
              9: Result := 'K6-III';
              else
                Result := 'K5/K6';
            end;
            6: Result := 'K7';
          else
            Result := 'AMD CPU';
        end;
      VIntel:
        case Info.Family of
          5:
            case Info.Model of
              0: Result := 'Pentium A-Step';
              1: Result := 'Pentium';
              2: Result := 'Pentium';
              3: Result := 'Pentium P24T OverDrive';
              4: Result := 'Pentium MMX';
              5: Result := 'Pentium P54C';
              7: Result := 'Pentium P55C 0.25 micron';
              8: Result := 'Pentium MMX mobile';
              else
                Result := 'Pentium';
            end;
          6:
            case Info.Model of
              0: Result := 'P6 A-Step';
              1: Result := 'Pentium Pro';
              3: Result := 'Pentium II';
              5: Result := 'Pentium II/Celeron';
              6: Result := 'Celeron A';
              7: Result := 'Pentium III/Xeon';
              8: Result := 'Pentium III E/Xeon';
              10: Result := 'Xeon A';
              else
                Result := 'Pentium Pro';
            end;
          else
            Result := 'Intel CPU';
        end;
      VCyrix:
        case Info.Family of
          5:
            case Info.Model of
              2: Result := '6x86/M1';
              4: Result := 'Media GX/GXm';
              else
                Result := '6x86/M1';
            end;
          6:
            case Info.Model of
              0: Result := '6x86MX/M2';
              5: Result := 'III';
              else
                Result := '6x86MX/M2';
            end;
          else
            Result := 'Cyrix CPU';
        end;
      VCentaur:
        case Info.Family of
          5:
            case Info.Model of
              4: Result := 'WinChip/C6';
              8: Result := 'WinChip 2/W2';
              9: Result := 'WinChip 3/W3';
              else
                Result := 'WinChip';
            end;
          6:
            case Info.Model of
              0: Result := 'C7';
              else
                Result := 'C7';
            end;
          else
            Result := 'Centaur/IDT CPU';
        end;
      else
        Result := 'Unknown CPU';
    end;
  end;

var
  Signature, Features: ULong;
  Str: PChar;
begin
  Info.Vendor := '';
  Info.Name := '';
  ZeroMemory(@Info, SizeOf(Info));
  if HasCPUID then
  begin
    CPUID(1, Signature, Features);
    Info.Family := (Signature shr 8) and $0F;
    Info.Model := (Signature shr 4) and $0F;
    Info.Stepping := Signature and $07;

    Info.Vendor := MakeVendor(GetCPUVendor);

    // Get AMD name string
    if (Info.Vendor = sAMD) and (Info.Family > 5) then
    begin
      Str := StrAlloc(48);
      GetStringAMD($80000002, Str);
      GetStringAMD($80000003, PChar(Long(Str) + 16));
      GetStringAMD($80000004, PChar(Long(Str) + 32));
      Info.Name := Str;
      StrDispose(Str);
    end
    else
      Info.Name := GetCPUName;

    // standard features
    if LongBool(Features and 1) then
      Info.Features := Info.Features + [cfFPU];
    if LongBool(Features and (1 shl 15)) then
      Info.Features := Info.Features + [cfCMOV];
    if LongBool(Features and (1 shl 23)) then
      Info.Features := Info.Features + [cfMMX];
    if LongBool(Features and (1 shl 25)) then
      Info.Features := Info.Features + [cfSSE];
    if LongBool(Features and (1 shl 26)) then
      Info.Features := Info.Features + [cfSSE2];

    // extended features
    CPUID($80000000, Signature, Features);
    if Signature > $80000000 then
    begin
      CPUID($80000001, Signature, Features);
      if LongBool(Features and (1 shl 22)) then
        Info.Features := Info.Features + [cfMMX2];
      if LongBool(Features and (1 shl 31)) then
        Info.Features := Info.Features + [cf3DNow];
      if LongBool(Features and (1 shl 30)) then
        Info.Features := Info.Features + [cf3DNowEx];
    end;
  end;
end;


end.

